@@ -130,9 +130,7 @@ CallbackCoro = Coroutine[Any, Any, Any]
 ConsumerCallback = Callable[[DeliveredMessage], CallbackCoro]
 ReturnCallback = Callable[[DeliveredMessage], Any]
 
-ArgumentsType = Dict[
-    str, Union[str, int, bool, Dict[str, Union[str, int, bool]]],
-]
+ArgumentsType = FieldTable
 
 ConfirmationFrameType = Union[
     Basic.Ack, Basic.Nack, Basic.Reject,
@@ -186,7 +184,7 @@ RpcReturnType = Optional[
 
 class ChannelFrame(NamedTuple):
     channel_number: int
-    frames: Iterable[Union[FrameType, Heartbeat]]
+    frames: Iterable[Union[FrameType, Heartbeat, ContentBody]]
     drain_future: Optional[asyncio.Future] = None
 
 
@@ -235,7 +233,7 @@ class AbstractBase(ABC):
 
     @abstractmethod
     async def close(
-        self, exc: Optional[ExceptionType] = asyncio.CancelledError()
+        self, exc: Optional[ExceptionType] = asyncio.CancelledError(),
     ) -> None:
         raise NotImplementedError
 
@@ -262,7 +260,7 @@ class AbstractChannel(AbstractBase):
     @abstractmethod
     async def basic_get(
         self, queue: str = "", no_ack: bool = False,
-        timeout: TimeoutType = None
+        timeout: TimeoutType = None,
     ) -> DeliveredMessage:
         raise NotImplementedError
 
@@ -281,8 +279,8 @@ class AbstractChannel(AbstractBase):
         *,
         no_ack: bool = False,
         exclusive: bool = False,
-        arguments: ArgumentsType = None,
-        consumer_tag: str = None,
+        arguments: Optional[ArgumentsType] = None,
+        consumer_tag: Optional[str] = None,
         timeout: TimeoutType = None
     ) -> spec.Basic.ConsumeOk:
         raise NotImplementedError
@@ -316,7 +314,7 @@ class AbstractChannel(AbstractBase):
         *,
         exchange: str = "",
         routing_key: str = "",
-        properties: spec.Basic.Properties = None,
+        properties: Optional[spec.Basic.Properties] = None,
         mandatory: bool = False,
         immediate: bool = False,
         timeout: TimeoutType = None
@@ -327,8 +325,8 @@ class AbstractChannel(AbstractBase):
     async def basic_qos(
         self,
         *,
-        prefetch_size: int = None,
-        prefetch_count: int = None,
+        prefetch_size: Optional[int] = None,
+        prefetch_count: Optional[int] = None,
         global_: bool = False,
         timeout: TimeoutType = None
     ) -> spec.Basic.QosOk:
@@ -352,7 +350,7 @@ class AbstractChannel(AbstractBase):
         auto_delete: bool = False,
         internal: bool = False,
         nowait: bool = False,
-        arguments: Optional[Dict[str, Any]] = None,
+        arguments: Optional[ArgumentsType] = None,
         timeout: TimeoutType = None
     ) -> spec.Exchange.DeclareOk:
         raise NotImplementedError
@@ -376,7 +374,7 @@ class AbstractChannel(AbstractBase):
         routing_key: str = "",
         *,
         nowait: bool = False,
-        arguments: dict = None,
+        arguments: Optional[ArgumentsType] = None,
         timeout: TimeoutType = None
     ) -> spec.Exchange.BindOk:
         raise NotImplementedError
@@ -389,7 +387,7 @@ class AbstractChannel(AbstractBase):
         routing_key: str = "",
         *,
         nowait: bool = False,
-        arguments: dict = None,
+        arguments: Optional[ArgumentsType] = None,
         timeout: TimeoutType = None
     ) -> spec.Exchange.UnbindOk:
         raise NotImplementedError
@@ -397,7 +395,7 @@ class AbstractChannel(AbstractBase):
     @abstractmethod
     async def flow(
         self, active: bool,
-        timeout: TimeoutType = None
+        timeout: TimeoutType = None,
     ) -> spec.Channel.FlowOk:
         raise NotImplementedError
 
@@ -408,8 +406,8 @@ class AbstractChannel(AbstractBase):
         exchange: str,
         routing_key: str = "",
         nowait: bool = False,
-        arguments: dict = None,
-        timeout: TimeoutType = None
+        arguments: Optional[ArgumentsType] = None,
+        timeout: TimeoutType = None,
     ) -> spec.Queue.BindOk:
         raise NotImplementedError
 
@@ -423,7 +421,7 @@ class AbstractChannel(AbstractBase):
         exclusive: bool = False,
         auto_delete: bool = False,
         nowait: bool = False,
-        arguments: dict = None,
+        arguments: Optional[ArgumentsType] = None,
         timeout: TimeoutType = None
     ) -> spec.Queue.DeclareOk:
         raise NotImplementedError
@@ -435,14 +433,14 @@ class AbstractChannel(AbstractBase):
         if_unused: bool = False,
         if_empty: bool = False,
         nowait: bool = False,
-        timeout: TimeoutType = None
+        timeout: TimeoutType = None,
     ) -> spec.Queue.DeleteOk:
         raise NotImplementedError
 
     @abstractmethod
     async def queue_purge(
         self, queue: str = "", nowait: bool = False,
-        timeout: TimeoutType = None
+        timeout: TimeoutType = None,
     ) -> spec.Queue.PurgeOk:
         raise NotImplementedError
 
@@ -452,20 +450,20 @@ class AbstractChannel(AbstractBase):
         queue: str = "",
         exchange: str = "",
         routing_key: str = "",
-        arguments: dict = None,
-        timeout: TimeoutType = None
+        arguments: Optional[ArgumentsType] = None,
+        timeout: TimeoutType = None,
     ) -> spec.Queue.UnbindOk:
         raise NotImplementedError
 
     @abstractmethod
     async def tx_commit(
-        self, timeout: TimeoutType = None
+        self, timeout: TimeoutType = None,
     ) -> spec.Tx.CommitOk:
         raise NotImplementedError
 
     @abstractmethod
     async def tx_rollback(
-        self, timeout: TimeoutType = None
+        self, timeout: TimeoutType = None,
     ) -> spec.Tx.RollbackOk:
         raise NotImplementedError
 
@@ -476,7 +474,7 @@ class AbstractChannel(AbstractBase):
     @abstractmethod
     async def confirm_delivery(
         self, nowait: bool = False,
-        timeout: TimeoutType = None
+        timeout: TimeoutType = None,
     ) -> spec.Confirm.SelectOk:
         raise NotImplementedError
 
@@ -513,7 +511,9 @@ class AbstractConnection(AbstractBase):
         raise NotImplementedError
 
     @abstractmethod
-    async def connect(self, client_properties: Dict[str, Any] = None) -> bool:
+    async def connect(
+        self, client_properties: Optional[FieldTable] = None,
+    ) -> bool:
         raise NotImplementedError
 
     @abstractproperty
@@ -538,7 +538,7 @@ class AbstractConnection(AbstractBase):
 
     async def channel(
         self,
-        channel_number: int = None,
+        channel_number: Optional[int] = None,
         publisher_confirms: bool = True,
         frame_buffer_size: int = FRAME_BUFFER_SIZE,
         timeout: TimeoutType = None,
@@ -563,6 +563,13 @@ class AbstractConnection(AbstractBase):
     async def ready(self) -> None:
         raise NotImplementedError
 
+    @abstractmethod
+    async def update_secret(
+        self, new_secret: str, *,
+        reason: str = "", timeout: TimeoutType = None,
+    ) -> spec.Connection.UpdateSecretOk:
+        raise NotImplementedError
+
 
 __all__ = (
     "AbstractBase", "AbstractChannel", "AbstractConnection",
