@@ -7,9 +7,9 @@ from weakref import WeakSet
 
 from .abc import (
     AbstractBase, AbstractFutureStore, CoroutineType, ExceptionType, TaskType,
-    TaskWrapper,
+    TaskWrapper, TimeoutType,
 )
-from .tools import shield
+from .tools import Countdown, shield
 
 
 T = TypeVar("T")
@@ -101,7 +101,7 @@ class Base(AbstractBase):
         return future
 
     def _cancel_tasks(
-        self, exc: ExceptionType = None,
+        self, exc: Optional[ExceptionType] = None,
     ) -> Coroutine[Any, Any, None]:
         return self.__future_store.reject_all(exc)
 
@@ -116,7 +116,7 @@ class Base(AbstractBase):
 
     @abc.abstractmethod
     async def _on_close(
-        self, exc: Optional[ExceptionType] = None
+        self, exc: Optional[ExceptionType] = None,
     ) -> None:  # pragma: no cover
         return
 
@@ -131,12 +131,14 @@ class Base(AbstractBase):
             await self._cancel_tasks(exc)
 
     async def close(
-        self, exc: Optional[ExceptionType] = asyncio.CancelledError
+        self, exc: Optional[ExceptionType] = asyncio.CancelledError,
+        timeout: TimeoutType = None,
     ) -> None:
         if self.is_closed:
             return None
 
-        await self.loop.create_task(self.__closer(exc))
+        countdown = Countdown(timeout)
+        await countdown(self.__closer(exc))
 
     def __repr__(self) -> str:
         cls_name = self.__class__.__name__
